//
// Copyright 2002,2003 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include <OPENR/core_macro.h>
#include "TCPReceptor.h"

TCPReceptor::TCPReceptor() : tcpReceptorState(TCPRS_IDLE),
			     connectionIndex(-1)
{
}

OStatus
TCPReceptor::DoInit(const OSystemEvent& event)
{
    OSYSDEBUG(("TCPReceptor::DoInit()\n"));

    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;

    return oSUCCESS;
}

OStatus
TCPReceptor::DoStart(const OSystemEvent& event)
{
    OSYSDEBUG(("TCPReceptor::DoStart()\n"));

    ipstackRef = antStackRef("IPStack");

    for (int index = 0; index < TCPRECEPTOR_CONNECTION_MAX; index++) {
        OStatus result = InitTCPConnection(index);
        if (result != oSUCCESS) return oFAIL;
    }

    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;

    tcpReceptorState = TCPRS_LISTEN;

    Listen(0);

    return oSUCCESS;
}

OStatus
TCPReceptor::DoStop(const OSystemEvent& event)
{
    OSYSDEBUG(("TCPReceptor::DoStop()\n"));

    tcpReceptorState = TCPRS_IDLE;

    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;

    return oSUCCESS;
}

OStatus
TCPReceptor::DoDestroy(const OSystemEvent& event)
{
    DELETE_ALL_SUBJECT_AND_OBSERVER;

    return oSUCCESS;
}

OStatus
TCPReceptor::Listen(int index)
{
    OSYSDEBUG(("TCPReceptor::Listen()\n"));

    if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    //
    // Create endpoint
    //
    antEnvCreateEndpointMsg tcpCreateMsg(EndpointType_TCP,
                                         TCPRECEPTOR_BUFFER_SIZE * 2);
    tcpCreateMsg.Call(ipstackRef, sizeof(tcpCreateMsg));
    if (tcpCreateMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "TCPReceptor::Listen()",
                  "Can't create endpoint",
                  index, tcpCreateMsg.error));
        return oFAIL;
    }
    connection[index].endpoint = tcpCreateMsg.moduleRef;

    //
    // Listen
    //
    TCPEndpointListenMsg listenMsg(connection[index].endpoint,
                                   IP_ADDR_ANY, TCPRECEPTOR_PORT);
    listenMsg.continuation = (void*)index;

    listenMsg.Send(ipstackRef, myOID_,
                   Extra_Entry[entryListenCont], sizeof(listenMsg));
    
    connection[index].state = CONNECTION_LISTENING;

    return oSUCCESS;
}

void
TCPReceptor::ListenCont(ANTENVMSG msg)
{
    OSYSDEBUG(("TCPReceptor::ListenCont()\n"));

    TCPEndpointListenMsg* listenMsg
        = (TCPEndpointListenMsg*)antEnvMsg::Receive(msg);
    int index = (int)listenMsg->continuation;

    if (listenMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "TCPReceptor::ListenCont()",
                  "FAILED. listenMsg->error", listenMsg->error));
        Close(index);
        return;
    }

    connection[index].state = CONNECTION_CONNECTED;

    SetReceiveData(index);
    Receive(index);
}

OStatus
TCPReceptor::Send(int index)
{
    OSYSDEBUG(("TCPReceptor::Send %d bytes\n", connection[index].sendSize ));

    if (connection[index].sendSize == 0 ||
        connection[index].state != CONNECTION_CONNECTED) return oFAIL;

    TCPEndpointSendMsg sendMsg(connection[index].endpoint,
                               connection[index].sendData,
                               connection[index].sendSize);
    sendMsg.continuation = (void*)index;

    sendMsg.Send(ipstackRef, myOID_,
                 Extra_Entry[entrySendCont],
                 sizeof(TCPEndpointSendMsg));

    connection[index].state = CONNECTION_SENDING;
    connection[index].sendSize = 0;
    return oSUCCESS;
}

void
TCPReceptor::SendCont(ANTENVMSG msg)
{
    OSYSDEBUG(("TCPReceptor::SendCont()\n"));

    TCPEndpointSendMsg* sendMsg = (TCPEndpointSendMsg*)antEnvMsg::Receive(msg);
    int index = (int)(sendMsg->continuation);

    if (sendMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "TCPReceptor::SendCont()",
                  "FAILED. sendMsg->error", sendMsg->error));
        Close(index);
        return;
    }

    //OSYSPRINT(("sent : %-6.6s\n", connection[index].sendData));

    connection[index].state = CONNECTION_CONNECTED;

    if ( connection[index].receiveNext ) {
	SetReceiveData(index);
	Receive(index);
    } else {
	Close(index);
    }
}

OStatus
TCPReceptor::Receive(int index)
{
    OSYSDEBUG(("TCPReceptor::Receive()\n"));

    if (connection[index].state != CONNECTION_CONNECTED &&
        connection[index].state != CONNECTION_SENDING) return oFAIL;

    TCPEndpointReceiveMsg receiveMsg(connection[index].endpoint,
                                     connection[index].recvData,
                                     1, connection[index].recvSize);
    receiveMsg.continuation = (void*)index;

    receiveMsg.Send(ipstackRef, myOID_,
                    Extra_Entry[entryReceiveCont], sizeof(receiveMsg));

    return oSUCCESS;
}

void
TCPReceptor::ReceiveCont(ANTENVMSG msg)
{
    OSYSDEBUG(("TCPReceptor::ReceiveCont()\n"));

    TCPEndpointReceiveMsg* receiveMsg
        = (TCPEndpointReceiveMsg*)antEnvMsg::Receive(msg);
    int index = (int)(receiveMsg->continuation);

    if (receiveMsg->error != TCP_SUCCESS) {
	OSYSDEBUG(( "Connection closed(%d)\n", receiveMsg->error ));
        Close(index);
        return;
    }

    connection[index].receiveNext = true;
    if ( receiveMsg->sizeMin < 12 ) {
	// ɬפʼХȿãƤʤä
	OSYSPRINT(( "SHORT(%d,%-16.16s)\n",
		    receiveMsg->sizeMin, connection[index].recvData ));

	connection[index].receiveNext = false;
	connection[index].sendSize = 21;
	memcpy( connection[index].sendData,
		"AIBO00000009*NG*SHORT",
		connection[index].sendSize );
	Send( index );
    } else if ( memcmp( connection[index].recvData, "AIBO", 4 ) ) {
	// Magicפʤä
	OSYSPRINT(( "MAGIC(%-16.16s)\n", connection[index].recvData ));

	connection[index].receiveNext = false;
	connection[index].sendSize = 24;
	memcpy( connection[index].sendData,
		"AIBO00000012*NG*BADMAGIC",
		connection[index].sendSize );
	Send( index );
    } else {
	// ʥåƤ
	char sizestr[9];
	memcpy( sizestr, &connection[index].recvData[4], 8 );
	sizestr[8] = '\0';
	int size = atoi(sizestr);

	OSYSPRINT(( "recvData : %-16.16s\n", connection[index].recvData ));
	OSYSPRINT(( "receiving : %d,%-10.10s\n",
		    size, &connection[index].recvData[4] ));

	ProcessReceivedMessage( index,
				&connection[index].recvData[12],
				atoi(sizestr) );
    }
}

OStatus
TCPReceptor::Close(int index)
{
    OSYSDEBUG(("TCPReceptor::Close()\n"));

    if (connection[index].state == CONNECTION_CLOSED ||
        connection[index].state == CONNECTION_CLOSING) return oFAIL;

    TCPEndpointCloseMsg closeMsg(connection[index].endpoint);
    closeMsg.continuation = (void*)index;

    closeMsg.Send(ipstackRef, myOID_,
                  Extra_Entry[entryCloseCont], sizeof(closeMsg));

    connection[index].state = CONNECTION_CLOSING;

    return oSUCCESS;
}

void
TCPReceptor::CloseCont(ANTENVMSG msg)
{
    OSYSDEBUG(("TCPReceptor::CloseCont()\n"));

    TCPEndpointCloseMsg* closeMsg
        = (TCPEndpointCloseMsg*)antEnvMsg::Receive(msg);
    int index = (int)(closeMsg->continuation);

    connection[index].state = CONNECTION_CLOSED;
    Listen(index);
}

OStatus
TCPReceptor::InitTCPConnection(int index)
{
    OSYSDEBUG(("TCPReceptor::InitTCPConnection()\n"));

    connection[index].state = CONNECTION_CLOSED;

    // 
    // Allocate send buffer
    //
    antEnvCreateSharedBufferMsg sendBufferMsg(TCPRECEPTOR_BUFFER_SIZE);

    sendBufferMsg.Call(ipstackRef, sizeof(sendBufferMsg));
    if (sendBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "TCPReceptor::InitTCPConnection()",
                  "Can't allocate send buffer",
                  index, sendBufferMsg.error));
        return oFAIL;
    }

    connection[index].sendBuffer = sendBufferMsg.buffer;
    connection[index].sendBuffer.Map();
    connection[index].sendData
        = (byte*)(connection[index].sendBuffer.GetAddress());

    //
    // Allocate receive buffer
    //
    antEnvCreateSharedBufferMsg recvBufferMsg(TCPRECEPTOR_BUFFER_SIZE);

    recvBufferMsg.Call(ipstackRef, sizeof(recvBufferMsg));
    if (recvBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "TCPReceptor::InitTCPConnection()",
                  "Can't allocate receive buffer",
                  index, recvBufferMsg.error));
        return oFAIL;
    }

    connection[index].recvBuffer = recvBufferMsg.buffer;
    connection[index].recvBuffer.Map();
    connection[index].recvData
        = (byte*)(connection[index].recvBuffer.GetAddress());

    return oSUCCESS;
}

void
TCPReceptor::SetSendData(int index)
{
    memcpy(connection[index].sendData,
           connection[index].recvData, connection[index].sendSize);
}

void
TCPReceptor::SetReceiveData(int index)
{
    connection[index].recvSize = TCPRECEPTOR_BUFFER_SIZE;
}
